import unittest
import warnings
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
import pandas as pd
import pytest
import regex
from lxml import etree

from dsp_tools.commands import excel2xml
from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import BaseError

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)
# ruff: noqa: PT027 (pytest-unittest-raises-assertion) (remove this line when pytest is used instead of unittest)


class TestPropsGeneral(unittest.TestCase):
    def run_test(
        self: unittest.TestCase,
        prop: str,
        method: Callable[..., etree._Element],
        different_values: Sequence[Union[str, int, float, bool]],
        invalid_values: Sequence[Any],
        listname: Optional[str] = None,
    ) -> None:
        """
        XML-properties have always a similar structure,
        and all make_*_prop() methods have some similar things to test.
        This method executes the tests in a parametrized way.

        Args:
            prop: the name of the property
            method: the make_*_prop() method
            different_values: some valid values
            invalid_values: some invalid values
            listname: to check the method make_list_prop, a list name is necessary

        Raises:
            BaseError: if make_list_prop() should be tested, but no listname is provided
        """
        identical_values = [different_values[0]] * 3
        maximum = len(different_values)

        # prepare the test cases of the form (expected_xml, kwargs for the method to generate XML)
        testcases: list[tuple[str, dict[str, Any]]] = []
        # pass every element of different_values separately
        for val in different_values:
            testcases.extend(
                [
                    (
                        f'<{prop}-prop name=":test"><{prop} permissions="open">{val}</{prop}></{prop}-prop>',
                        dict(name=":test", value=val),
                    ),
                    (
                        f'<{prop}-prop name=":test"><{prop} permissions="restricted">{val}</{prop}></{prop}-prop>',
                        dict(name=":test", value=excel2xml.PropertyElement(val, permissions="restricted")),
                    ),
                    (
                        f'<{prop}-prop name=":test"><{prop} permissions="restricted" comment="comment">{val}'
                        f"</{prop}></{prop}-prop>",
                        dict(
                            name=":test",
                            value=excel2xml.PropertyElement(val, permissions="restricted", comment="comment"),
                        ),
                    ),
                ]
            )
        # pass the elements of different_values group-wise
        testcases.extend(
            [
                (
                    f'<{prop}-prop name=":test">'
                    f'<{prop} permissions="open">{identical_values[0]}</{prop}>'
                    f'<{prop} permissions="open">{identical_values[1]}</{prop}>'
                    f'<{prop} permissions="open">{identical_values[2]}</{prop}>'
                    f"</{prop}-prop>",
                    dict(name=":test", value=identical_values),
                ),
                (
                    f'<{prop}-prop name=":test">'
                    f'<{prop} permissions="open">{different_values[0 % maximum]}</{prop}>'
                    f'<{prop} permissions="open">{different_values[1 % maximum]}</{prop}>'
                    f'<{prop} permissions="open">{different_values[2 % maximum]}</{prop}>'
                    f"</{prop}-prop>",
                    dict(
                        name=":test",
                        value=[
                            different_values[0 % maximum],
                            different_values[1 % maximum],
                            different_values[2 % maximum],
                        ],
                    ),
                ),
                (
                    f'<{prop}-prop name=":test">'
                    f'<{prop} permissions="restricted" comment="comment1">{different_values[3 % maximum]}</{prop}>'
                    f'<{prop} permissions="open" comment="comment2">{different_values[4 % maximum]}</{prop}>'
                    f'<{prop} permissions="restricted" comment="comment3">{different_values[5 % maximum]}</{prop}>'
                    f"</{prop}-prop>",
                    dict(
                        name=":test",
                        value=[
                            excel2xml.PropertyElement(
                                different_values[3 % maximum], permissions="restricted", comment="comment1"
                            ),
                            excel2xml.PropertyElement(
                                different_values[4 % maximum], permissions="open", comment="comment2"
                            ),
                            excel2xml.PropertyElement(
                                different_values[5 % maximum], permissions="restricted", comment="comment3"
                            ),
                        ],
                    ),
                ),
            ]
        )

        # run the test cases
        for tc in testcases:
            xml_expected = tc[0]
            kwargs_to_generate_xml = tc[1]
            if prop == "list":
                # a <list-prop> has the additional attribute list="listname"
                xml_expected = regex.sub(r"<list-prop", f'<list-prop list="{listname}"', xml_expected)
                kwargs_to_generate_xml["list_name"] = listname
            elif prop == "text":
                # a <text> has the additional attribute encoding="utf8"
                # (the other encoding, xml, is tested in the caller)
                xml_expected = regex.sub(
                    r"<text (permissions=\".+?\")( comment=\".+?\")?",
                    '<text \\1\\2 encoding="utf8"',
                    xml_expected,
                )
            xml_returned_as_element = method(**kwargs_to_generate_xml)
            xml_returned = etree.tostring(xml_returned_as_element, encoding="unicode")
            xml_returned = regex.sub(
                r" xmlns(:.+?)?=\".+?\"", "", xml_returned
            )  # remove all xml namespace declarations
            self.assertEqual(
                xml_expected, xml_returned, msg=f"Method {method.__name__} failed with kwargs {kwargs_to_generate_xml}"
            )

        # perform illegal actions
        # pass invalid values as param "value"
        for invalid_value in invalid_values:
            kwargs_invalid_value = dict(name=":test", value=invalid_value)
            if prop == "list":
                if not listname:
                    raise BaseError("listname must be set to test make_list_prop()")
                kwargs_invalid_value["list_name"] = listname
            with self.assertWarns(Warning, msg=f"Method {method.__name__} failed with kwargs {kwargs_invalid_value}"):
                method(**kwargs_invalid_value)

    def test_make_color_prop(self) -> None:
        prop = "color"
        method = excel2xml.make_color_prop
        different_values = ["#012345", "#abcdef", "#0B0B0B", "#AAAAAA", "#1a2b3c"]
        invalid_values = ["#0000000", "#00000G"]
        self.run_test(prop, method, different_values, invalid_values)

    def test_make_date_prop(self) -> None:
        prop = "date"
        method = excel2xml.make_date_prop
        different_values = [
            "CE:1849:CE:1850",
            "GREGORIAN:1848-01:1849-02",
            "2022",
            "GREGORIAN:CE:0476-09-04:CE:0476-09-04",
            "GREGORIAN:CE:2014-01-31",
            "JULIAN:BC:1:AD:200",
        ]
        invalid_values = ["GREGORIAN:CE:0476-09-010:CE:0476-09-04"]
        self.run_test(prop, method, different_values, invalid_values)

    def test_make_decimal_prop(self) -> None:
        prop = "decimal"
        method = excel2xml.make_decimal_prop
        different_values: list[Union[str, float, int]] = ["3.14159", 3.14159, "1.3e3", "100", ".1", 100]
        invalid_values = ["string"]
        self.run_test(prop, method, [float(x) for x in different_values], invalid_values)

    def test_make_geometry_prop(self) -> None:
        prop = "geometry"
        method = excel2xml.make_geometry_prop
        different_values = [
            '{"type": "rectangle", "lineWidth": 2, '
            '"points": [{"x": 0.08, "y": 0.16}, {"x": 0.73, "y": 0.72}], "original_index": 0}',
            '{"type": "rectangle", "lineWidth": 1, '
            '"points": [{"x": 0.10, "y": 0.10}, {"x": 0.10, "y": 0.10}], "original_index": 1}',
        ]
        invalid_values = ["100", 100, [0], '{"type": "polygon"}']
        self.run_test(prop, method, different_values, invalid_values)

    def test_make_geoname_prop(self) -> None:
        prop = "geoname"
        method = excel2xml.make_geoname_prop
        different_values: list[Union[int, str]] = [1283416, "1283416", 71, "71", 10000000, "10000000"]
        invalid_values = ["text", 10.0, ["text"]]
        self.run_test(prop, method, different_values, invalid_values)

    def test_make_integer_prop(self) -> None:
        prop = "integer"
        method = excel2xml.make_integer_prop
        different_values: list[Union[int, str, float]] = [1283416, "1283416", 3.14159, " 11 ", 0, "0"]
        invalid_values = [" 10.3 ", "text", ["text"]]
        self.run_test(prop, method, [int(x) for x in different_values], invalid_values)

    def test_make_list_prop(self) -> None:
        prop = "list"
        method = excel2xml.make_list_prop
        different_values = ["first-node", "second-node", "third-node", "fourth-node", "fifth-node"]
        invalid_values = [10.0]
        self.run_test(prop, method, different_values, invalid_values, ":myList")

    def test_make_resptr_prop(self) -> None:
        prop = "resptr"
        method = excel2xml.make_resptr_prop
        different_values = ["resource_1", "resource_2", "resource_3", "resource_4", "resource_5"]
        invalid_values = [True, 10.0, 5]
        self.run_test(prop, method, different_values, invalid_values)

    def test_make_time_prop(self) -> None:
        prop = "time"
        method = excel2xml.make_time_prop
        different_values = [
            "2019-10-23T13:45:12.01-14:00",
            "2019-10-23T13:45:12-14:00",
            "2019-10-23T13:45:12Z",
            "2019-10-23T13:45:12-13:30",
            "2019-10-23T13:45:12+01:00",
            "2019-10-23T13:45:12.1111111+01:00",
            "2019-10-23T13:45:12.123456789012Z",
        ]
        invalid_values = [
            True,
            10.0,
            5,
            "2019-10-2",
            "CE:1849:CE:1850",
            "2019-10-23T13:45:12.1234567890123Z",
            "2022",
            "GREGORIAN:CE:2014-01-31",
        ]
        self.run_test(prop, method, different_values, invalid_values)

    def test_make_uri_prop(self) -> None:
        prop = "uri"
        method = excel2xml.make_uri_prop
        different_values = [
            "https://www.test-case.ch/",
            "https://reg-exr.com:3000",
            "https://reg-exr.com:3000/path/to/file",
            "https://reg-exr.com:3000/path/to/file#fragment,fragment",
            "https://reg-exr.com:3000/path/to/file?query=test",
            "https://reg-exr.com:3000/path/to/file?query=test#fragment",
            "https://reg-exr.com/path/to/file?query=test#fragment",
            "http://www.168.1.1.0/path",
            "http://www.168.1.1.0:4200/path",
            "http://[2001:0db8:0000:0000:0000:8a2e:0370:7334]:4200/path",
            "https://en.wikipedia.org/wiki/Haiku#/media/File:Basho_Horohoroto.jpg",
        ]
        invalid_values = ["https:", 10.0, 5, "www.test.com"]
        self.run_test(prop, method, different_values, invalid_values)

    @pytest.mark.filterwarnings("ignore::dsp_tools.models.custom_warnings.DspToolsUserWarning")
    def test_make_text_prop(self) -> None:
        prop = "text"
        method = excel2xml.make_text_prop
        different_values = ["text_1", " ", "!", "?", "-", "_", "None"]
        invalid_values = [True, 10.0, 5, ""]
        self.run_test(prop, method, different_values, invalid_values)
        self.assertRaises(
            BaseError,
            lambda: excel2xml.make_text_prop(":test", excel2xml.PropertyElement(value="a", encoding="unicode")),
        )

    def test_make_boolean_prop(self) -> None:
        # prepare true_values
        true_values_orig: list[Union[bool, str, int]] = [True, "TRue", "TruE", "1", 1, "yes", "YES", "yEs"]
        true_values_as_propelem = [excel2xml.PropertyElement(x) for x in true_values_orig]
        true_values = true_values_orig + true_values_as_propelem

        # prepare false_values
        false_values_orig: list[Union[bool, str, int]] = [False, "false", "False", "falSE", "0", 0, "no", "No", "nO"]
        false_values_as_propelem = [excel2xml.PropertyElement(x) for x in false_values_orig]
        false_values = false_values_orig + false_values_as_propelem

        unsupported_values: list[Any] = [
            np.nan,
            pd.NA,
            "N/A",
            "NA",
            "na",
            "None",
            "",
            " ",
            "-",
            None,
            [True, False],
            [0, 0, 1],
            ["True", "false"],
        ]

        true_xml_expected = '<boolean-prop name=":test"><boolean permissions="open">true</boolean></boolean-prop>'
        false_xml_expected = '<boolean-prop name=":test"><boolean permissions="open">false</boolean></boolean-prop>'

        for true_value in true_values:
            true_xml = etree.tostring(excel2xml.make_boolean_prop(":test", true_value), encoding="unicode")
            true_xml = regex.sub(r" xmlns(:.+?)?=\".+?\"", "", true_xml)
            self.assertEqual(true_xml, true_xml_expected, msg=f"Failed with '{true_value}'")
        for false_value in false_values:
            false_xml = etree.tostring(excel2xml.make_boolean_prop(":test", false_value), encoding="unicode")
            false_xml = regex.sub(r" xmlns(:.+?)?=\".+?\"", "", false_xml)
            self.assertEqual(false_xml, false_xml_expected, msg=f"Failed with '{false_value}'")
        for unsupported_value in unsupported_values:
            with self.assertRaises(BaseError):
                excel2xml.make_boolean_prop(":test", unsupported_value)


class TestTextProp:
    def test_make_text_prop_utf8_lt_gt_amp(self) -> None:
        original = "1 < 2 & 4 > 3"
        expected = "1 &lt; 2 &amp; 4 &gt; 3"
        returned = etree.tostring(excel2xml.make_text_prop(":test", original), encoding="unicode")
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_utf8_pseudo_tag(self) -> None:
        original = "txt <txt>txt</txt> txt"
        expected = "txt &lt;txt&gt;txt&lt;/txt&gt; txt"
        returned = etree.tostring(excel2xml.make_text_prop(":test", original), encoding="unicode")
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_utf8_escape(self) -> None:
        original = "txt &amp; txt"
        expected = "txt &amp;amp; txt"
        returned = etree.tostring(excel2xml.make_text_prop(":test", original), encoding="unicode")
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_xml_standard_standoff_tag(self) -> None:
        original = "text <strong>and</strong> text"
        expected = "text <strong>and</strong> text"
        returned = etree.tostring(
            excel2xml.make_text_prop(":test", excel2xml.PropertyElement(original, encoding="xml")), encoding="unicode"
        )
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_xml_unsupported_tag(self) -> None:
        original = "text <unsupported>tag</unsupported> text"
        expected = "text &lt;unsupported&gt;tag&lt;/unsupported&gt; text"
        returned = etree.tostring(
            excel2xml.make_text_prop(":test", excel2xml.PropertyElement(original, encoding="xml")), encoding="unicode"
        )
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_xml_salsah_link(self) -> None:
        original = 'a <a class="salsah-link" href="IRI:test_thing_0:IRI">link</a> text'
        expected = 'a <a class="salsah-link" href="IRI:test_thing_0:IRI">link</a> text'
        returned = etree.tostring(
            excel2xml.make_text_prop(":test", excel2xml.PropertyElement(original, encoding="xml")), encoding="unicode"
        )
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_xml_escaped_tag(self) -> None:
        original = "text <strong>and</strong> text"
        expected = "text <strong>and</strong> text"
        returned = etree.tostring(
            excel2xml.make_text_prop(":test", excel2xml.PropertyElement(original, encoding="xml")), encoding="unicode"
        )
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_xml_lt_gt_amp(self) -> None:
        original = "1 < 2 & 4 > 3"
        expected = "1 &lt; 2 &amp; 4 &gt; 3"
        returned = etree.tostring(
            excel2xml.make_text_prop(":test", excel2xml.PropertyElement(original, encoding="xml")), encoding="unicode"
        )
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_xml_escape_sequence(self) -> None:
        original = "text &amp; text"
        expected = "text &amp; text"
        returned = etree.tostring(
            excel2xml.make_text_prop(":test", excel2xml.PropertyElement(original, encoding="xml")), encoding="unicode"
        )
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_special_characters_xml(self) -> None:
        originals = ["'uuas\\. 11` \\a\\ i! 1 ?7", "Rinne   \\Rinne"]
        for orig in originals:
            returned = etree.tostring(
                excel2xml.make_text_prop(":test", excel2xml.PropertyElement(orig, encoding="xml")), encoding="unicode"
            )
            returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
            assert returned == orig

    def test_make_text_prop_special_characters_utf8(self) -> None:
        originals = ["'uuas\\. 11` \\a\\ i! 1 ?7", "Rinne   \\Rinne"]
        for orig in originals:
            returned = etree.tostring(excel2xml.make_text_prop(":test", orig), encoding="unicode")
            returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
            assert returned == orig

    def test_make_text_prop_disallowed_named_char_refs(self) -> None:
        # the following will be displayed as "ä €" in the XML and in DSP-APP  # noqa: RUF003 (ambiguous character)
        original = "<p>text &auml;&nbsp;&euro; text</p>"
        expected = "<p>text ä\xa0€ text</p>"
        prop = excel2xml.make_text_prop(":test", excel2xml.PropertyElement(original, encoding="xml"))
        returned = etree.tostring(prop, encoding="unicode")
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected

    def test_make_text_prop_allowed_named_char_refs(self) -> None:
        original = "<p>text &lt; &amp; &gt; text</p>"
        expected = "<p>text &lt; &amp; &gt; text</p>"
        prop = excel2xml.make_text_prop(":test", excel2xml.PropertyElement(original, encoding="xml"))
        returned = etree.tostring(prop, encoding="unicode")
        returned = regex.sub(r"</?text(-prop)?( [^>]+)?>", "", returned)
        assert returned == expected


class TestBitstreamProp:
    def test_make_bitstream_prop_from_string(self) -> None:
        res = excel2xml.make_bitstream_prop("foo/bar/baz.txt")
        assert str(res.tag).endswith("bitstream")
        assert res.attrib["permissions"] == "open"
        assert res.text == "foo/bar/baz.txt"

    def test_make_bitstream_prop_from_path(self) -> None:
        res = excel2xml.make_bitstream_prop(Path("foo/bar/baz.txt"))
        assert str(res.tag).endswith("bitstream")
        assert res.attrib["permissions"] == "open"
        assert res.text == "foo/bar/baz.txt"

    def test_make_bitstream_prop_custom_permissions(self) -> None:
        res = excel2xml.make_bitstream_prop("foo/bar/baz.txt", "restricted")
        assert str(res.tag).endswith("bitstream")
        assert res.attrib["permissions"] == "restricted"
        assert res.text == "foo/bar/baz.txt"

    def test_make_bitstream_prop_valid_file(self) -> None:
        with warnings.catch_warnings():
            warnings.filterwarnings("error", category=UserWarning)
            try:
                res = excel2xml.make_bitstream_prop("testdata/bitstreams/test.jpg", check=True)
            except UserWarning as e:
                raise AssertionError from e
        assert str(res.tag).endswith("bitstream")
        assert res.attrib["permissions"] == "open"
        assert res.text == "testdata/bitstreams/test.jpg"

    def test_make_bitstream_prop_invalid_file(self) -> None:
        with pytest.warns(DspToolsUserWarning, match=".*Failed validation in bitstream tag.*"):
            res = excel2xml.make_bitstream_prop("foo/bar/baz.txt", check=True)
        assert str(res.tag).endswith("bitstream")
        assert res.attrib["permissions"] == "open"
        assert res.text == "foo/bar/baz.txt"


@pytest.mark.parametrize(
    ("tag", "func"),
    [("isSegmentOf", excel2xml.make_isSegmentOf_prop), ("relatesTo", excel2xml.make_relatesTo_prop)],
)
class Test_isSegmentOf_and_relatesTo_Prop:
    def test_defaults(self, tag: str, func: Callable[..., etree._Element]) -> None:
        res = func("target_id")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.text == "target_id"

    def test_custom_params(self, tag: str, func: Callable[..., etree._Element]) -> None:
        res = func("target_id", "restricted", "my comment")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "restricted"
        assert res.attrib["comment"] == "my comment"
        assert res.text == "target_id"

    def test_invalid(self, tag: str, func: Callable[..., etree._Element]) -> None:
        with pytest.warns(DspToolsUserWarning):
            res = func("<NA>")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.text == "<NA>"


class Test_hasSegmentBounds_Prop:
    def test_defaults(self) -> None:
        res = excel2xml.make_hasSegmentBounds_prop(100, 200)
        assert str(res.tag).endswith("hasSegmentBounds")
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.attrib["segment_start"] == "100"
        assert res.attrib["segment_end"] == "200"
        assert res.text is None

    def test_custom_params(self) -> None:
        res = excel2xml.make_hasSegmentBounds_prop(10, 20, "restricted", "my comment")
        assert str(res.tag).endswith("hasSegmentBounds")
        assert res.attrib["permissions"] == "restricted"
        assert res.attrib["comment"] == "my comment"
        assert res.attrib["segment_start"] == "10"
        assert res.attrib["segment_end"] == "20"
        assert res.text is None

    def test_floats(self) -> None:
        res = excel2xml.make_hasSegmentBounds_prop(1.2, 3.4)
        assert str(res.tag).endswith("hasSegmentBounds")
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.attrib["segment_start"] == "1.2"
        assert res.attrib["segment_end"] == "3.4"
        assert res.text is None

    def test_nums_as_strings(self) -> None:
        res = excel2xml.make_hasSegmentBounds_prop("1.2", "3")  # type: ignore[arg-type]
        assert str(res.tag).endswith("hasSegmentBounds")
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.attrib["segment_start"] == "1.2"
        assert res.attrib["segment_end"] == "3.0"  # silent conversion to float is okay, since the db stores it that way
        assert res.text is None

    def test_start_less_than_end(self) -> None:
        with pytest.warns(DspToolsUserWarning, match="must be less than"):
            res = excel2xml.make_hasSegmentBounds_prop(5, 3)
        assert str(res.tag).endswith("hasSegmentBounds")
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.attrib["segment_start"] == "5"
        assert res.attrib["segment_end"] == "3"
        assert res.text is None

    def test_not_a_number(self) -> None:
        with pytest.warns(DspToolsUserWarning, match="must be integers or floats"):
            res = excel2xml.make_hasSegmentBounds_prop(segment_start="foo", segment_end=2)  # type: ignore[arg-type]
        assert str(res.tag).endswith("hasSegmentBounds")
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.attrib["segment_start"] == "foo"
        assert res.attrib["segment_end"] == "2"
        assert res.text is None


@pytest.mark.parametrize(
    ("tag", "func"),
    [("hasTitle", excel2xml.make_hasTitle_prop), ("hasKeyword", excel2xml.make_hasKeyword_prop)],
)
class Test_hasTitle_hasKeyword:
    def test_defaults(self, tag: str, func: Callable[..., etree._Element]) -> None:
        res = func("my text ...")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.text == "my text ..."

    def test_custom_params(self, tag: str, func: Callable[..., etree._Element]) -> None:
        res = func("my text ...", "restricted", "my comment")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "restricted"
        assert res.attrib["comment"] == "my comment"
        assert res.text == "my text ..."

    def test_invalid(self, tag: str, func: Callable[..., etree._Element]) -> None:
        with pytest.warns(DspToolsUserWarning):
            res = func("<NA>")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.text == "<NA>"


@pytest.mark.parametrize(
    ("tag", "func"),
    [("hasComment", excel2xml.make_hasComment_prop), ("hasDescription", excel2xml.make_hasDescription_prop)],
)
class Test_hasComment_hasDescription:
    def test_defaults(self, tag: str, func: Callable[..., etree._Element]) -> None:
        res = func("my text ...")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.text == "my text ..."

    def test_custom_params(self, tag: str, func: Callable[..., etree._Element]) -> None:
        res = func("my text ...", "restricted", "my comment")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "restricted"
        assert res.attrib["comment"] == "my comment"
        assert res.text == "my text ..."

    def test_invalid(self, tag: str, func: Callable[..., etree._Element]) -> None:
        with pytest.warns(DspToolsUserWarning):
            res = func("<NA>")
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        assert res.text == "<NA>"

    def test_richtext(self, tag: str, func: Callable[..., etree._Element]) -> None:
        text = "<p>my <strong>bold <em>and italiziced</em></strong> text ...</p>"
        res = func(text)
        assert str(res.tag).endswith(tag)
        assert res.attrib["permissions"] == "open"
        assert "comment" not in res.attrib
        serialized_text = regex.sub(rf"<ns0:{tag} .+?>|</ns0:{tag}>", "", etree.tostring(res, encoding="unicode"))
        assert serialized_text == text

    def test_invalid_richtext(self, tag: str, func: Callable[..., etree._Element]) -> None:  # noqa: ARG002 (unused arg)
        text = "<p> my <pseudo> & xml </>"
        with pytest.raises(BaseError, match=regex.escape("must be well-formed")):
            func(text)



import pytest
from requests.exceptions import RequestException

from dsp_tools.commands.xmlupload.models.input_problems import AllIIIFUriProblems
from dsp_tools.commands.xmlupload.models.input_problems import IIIFUriProblem


@pytest.fixture
def iiif_exception() -> IIIFUriProblem:
    return IIIFUriProblem(
        uri="http://www.example.org/",
        regex_has_passed=False,
        raised_exception_name="RequestException",
        original_text=str(RequestException("This is the request exception.")),
    )


@pytest.fixture
def iiif_uri_problem_ok_regex() -> IIIFUriProblem:
    return IIIFUriProblem(
        uri="https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2",
        regex_has_passed=True,
        status_code=404,
        original_text="This is the response text.",
    )


@pytest.fixture
def iiif_uri_problem_bad_regex() -> IIIFUriProblem:
    return IIIFUriProblem(
        uri="http://www.example.org/",
        regex_has_passed=False,
        status_code=404,
        original_text="This is the response text.",
    )


@pytest.fixture
def iiif_uri_problem_bad_regex_good_status_code() -> IIIFUriProblem:
    return IIIFUriProblem(
        uri="http://www.example.org/",
        regex_has_passed=False,
        status_code=200,
        original_text="This is the response text.",
    )


def test_iiif_uri_problem_ok_regex(iiif_uri_problem_ok_regex: IIIFUriProblem) -> None:
    assert iiif_uri_problem_ok_regex.get_msg() == (
        "URI: https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2\n"
        "    - Passed the internal regex check.\n"
        "    - The server did not respond as expected.\n"
        "    - Status code: 404\n"
        "    - Response text: This is the response text."
    )


def test_iiif_uri_problem_bad_regex(iiif_uri_problem_bad_regex: IIIFUriProblem) -> None:
    assert iiif_uri_problem_bad_regex.get_msg() == (
        "URI: http://www.example.org/\n"
        "    - Did not pass the internal regex check.\n"
        "    - The server did not respond as expected.\n"
        "    - Status code: 404\n"
        "    - Response text: This is the response text."
    )


def test_iiif_uri_problem_exception(iiif_exception: IIIFUriProblem) -> None:
    assert iiif_exception.get_msg() == (
        "URI: http://www.example.org/\n"
        "    - Did not pass the internal regex check.\n"
        "    - An error occurred during the network call: RequestException\n"
        "    - Error message: This is the request exception."
    )


def test_failing_regex(iiif_uri_problem_bad_regex_good_status_code: IIIFUriProblem) -> None:
    assert iiif_uri_problem_bad_regex_good_status_code.get_msg() == (
        "URI: http://www.example.org/\n"
        "    - Did not pass the internal regex check.\n"
        "    - The URI is correct and the server responded as expected.\n"
        "    - Please contact the dsp-tools development team with this information."
    )


def test_all(iiif_exception: IIIFUriProblem, iiif_uri_problem_ok_regex: IIIFUriProblem) -> None:
    problem = AllIIIFUriProblems(problems=[iiif_exception, iiif_uri_problem_ok_regex])
    assert problem.get_msg() == (
        "There were problems with the following IIIF URI(s):\n"
        "----------------------------\n"
        "URI: http://www.example.org/\n"
        "    - Did not pass the internal regex check.\n"
        "    - An error occurred during the network call: RequestException\n"
        "    - Error message: This is the request exception."
        "\n----------------------------\n"
        "URI: https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2\n"
        "    - Passed the internal regex check.\n"
        "    - The server did not respond as expected.\n"
        "    - Status code: 404\n"
        "    - Response text: This is the response text."
    )


if __name__ == "__main__":
    pytest.main([__file__])



from typing import Any

import pytest


@pytest.fixture
def response_all_list_one_project() -> dict[str, Any]:
    return {
        "lists": [
            {
                "id": "http://rdfh.ch/lists/9999/list1",
                "projectIri": "http://rdfh.ch/projects/projectIRI",
                "name": "firstList",
                "labels": [{"value": "List 1", "language": "en"}],
                "comments": [{"value": "This is the first list", "language": "en"}],
                "isRootNode": True,
            },
            {
                "id": "http://rdfh.ch/lists/9999/list2",
                "projectIri": "http://rdfh.ch/projects/projectIRI",
                "name": "secondList",
                "labels": [{"value": "List", "language": "en"}],
                "comments": [{"value": "This is the second list", "language": "en"}],
                "isRootNode": True,
            },
        ]
    }


@pytest.fixture
def response_all_list_one_project_no_lists() -> dict[str, Any]:
    return {"lists": []}


@pytest.fixture
def response_one_list() -> dict[str, Any]:
    return {
        "type": "ListGetResponseADM",
        "list": {
            "listinfo": {
                "id": "http://rdfh.ch/lists/9999/list1",
                "projectIri": "http://rdfh.ch/projects/projectIRI",
                "name": "firstList",
                "labels": [{"value": "List 1", "language": "en"}],
                "comments": [{"value": "This is the first list", "language": "en"}],
                "isRootNode": True,
            },
            "children": [
                {
                    "id": "http://rdfh.ch/lists/9999/n1",
                    "name": "n1",
                    "labels": [{"value": "Node 1", "language": "en"}],
                    "comments": [],
                    "position": 0,
                    "hasRootNode": "http://rdfh.ch/lists/9999/list1",
                    "children": [
                        {
                            "id": "http://rdfh.ch/lists/9999/n11",
                            "name": "n1.1",
                            "labels": [{"value": "Node 1.1", "language": "en"}],
                            "comments": [],
                            "position": 0,
                            "hasRootNode": "http://rdfh.ch/lists/9999/list1",
                            "children": [
                                {
                                    "id": "http://rdfh.ch/lists/9999/n111",
                                    "name": "n1.1.1",
                                    "labels": [{"value": "Node 1.1.1", "language": "en"}],
                                    "comments": [],
                                    "position": 0,
                                    "hasRootNode": "http://rdfh.ch/lists/9999/list1",
                                    "children": [],
                                },
                                {
                                    "id": "http://rdfh.ch/lists/9999/n112",
                                    "name": "n1.1.2",
                                    "labels": [{"value": "Node 1.1.2", "language": "en"}],
                                    "comments": [],
                                    "position": 1,
                                    "hasRootNode": "http://rdfh.ch/lists/9999/list1",
                                    "children": [],
                                },
                            ],
                        }
                    ],
                }
            ],
        },
    }



import warnings
from typing import Any

import pandas as pd
import regex
from lxml import etree
from namedentities.core import numeric_entities  # type: ignore[import-untyped]

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.xmllib.models.problems import IllegalTagProblem


def is_nonempty_value(value: Any) -> bool:
    """
    Check if a value is not empty.

    Args:
        value: value to check

    Returns:
        True if it is not empty

    Examples:
        ```python
        result = xmllib.is_nonempty_value("not empty")
        # result == True
        ```

        ```python
        result = xmllib.is_nonempty_value("")
        # result == False
        ```

        ```python
        result = xmllib.is_nonempty_value(None)
        # result == False
        ```
    """
    if isinstance(value, str) and len(value) == 0:
        return False
    return not pd.isna(value)


def is_bool_like(value: Any) -> bool:
    """
    Checks if a value is a bool or can be converted into a bool.
    It is case-insensitive, meaning that the words can also be capitalised.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_bool_like("yes")
        # result == True
        ```

        ```python
        result = xmllib.is_bool_like("not like a bool")
        # result == False
        ```
    """
    value = str(value).lower().strip()
    if value in ("false", "0", "0.0", "no", "non", "nein"):
        return True
    elif value in ("true", "1", "1.0", "yes", "oui", "ja"):
        return True
    return False


def is_color(value: Any) -> bool:
    """
    Checks if a value is a color value.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_color("#00ff66")
        # result == True
        ```

        ```python
        result = xmllib.is_color("not a color")
        # result == False
        ```
    """
    return bool(regex.search(r"^#[0-9a-f]{6}$", str(value).strip(), flags=regex.IGNORECASE))


def is_date(value: Any) -> bool:
    """
    Checks if a value is a date value.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_date("GREGORIAN:CE:2014-01-31:CE:2014-01-31")
        # result == True
        ```

        ```python
        result = xmllib.is_date("not a date")
        # result == False
        ```
    """
    calendar_optional = r"((GREGORIAN|JULIAN|ISLAMIC):)?"
    first_era_optional = r"((CE|BCE|BC|AD):)?"
    second_area_optional = r"(:(CE|BCE|BC|AD))?"
    date = r"\d{1,4}(?:-\d{1,2}){0,2}"
    date_mandatory = rf"({date})"
    date_optional = rf"(:{date})?"
    full_date_pattern = (
        rf"^{calendar_optional}{first_era_optional}{date_mandatory}{second_area_optional}{date_optional}$"
    )
    return bool(regex.search(full_date_pattern, str(value)))


def is_geoname(value: Any) -> bool:
    """
    Checks if a value is a geoname value.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_geoname("8879000")
        # result == True
        ```

        ```python
        result = xmllib.is_geoname("not a geoname code")
        # result == False
        ```
    """
    return is_integer(value)


def is_decimal(value: Any) -> bool:
    """
    Checks if a value is a float, an integer, or a string which can be converted into a float.

    Args:
        value: value to check

    Returns:
        True if conforms to the above-mentioned criteria.

    Examples:
        ```python
        result = xmllib.is_decimal("0.1")
        # result == True
        ```

        ```python
        # because this is equivalent to 9.0 it is accepted

        result = xmllib.is_decimal(9)
        # result == True
        ```

        ```python
        result = xmllib.is_decimal("not a decimal")
        # result == False
        ```
    """
    if pd.isna(value):
        return False

    match value:
        case bool():
            return False
        case int() | float():
            return True
    try:
        float(value)
        return True
    except ValueError:
        return False


def is_integer(value: Any) -> bool:
    """
    Checks if a value is an integer or a string which can be converted into an integer.

    Args:
        value: value to check

    Returns:
        True if conforms to the above-mentioned criteria.

    Examples:
        ```python
        result = xmllib.is_integer("1")
        # result == True
        ```

        ```python
        result = xmllib.is_integer(9.1)
        # result == False
        ```

        ```python
        result = xmllib.is_integer("not an integer")
        # result == False
        ```
    """
    match value:
        case bool():
            return False
        case int():
            return True
        case str():
            return bool(regex.search(r"^\d+$", value))
        case _:
            return False


def is_string_like(value: Any) -> bool:
    """
    Checks if a value is a string.

    Args:
        value: value to check

    Returns:
        True if it is a string

    Examples:
        ```python
        result = xmllib.is_string_like("this is a string")
        # result == True
        ```

        ```python
        # because numbers, floats, etc. can be converted to strings they are accepted

        result = xmllib.is_string_like(1)
        # result == True
        ```

        ```python
        result = xmllib.is_string_like(None)
        # result == False
        ```
    """
    if pd.isna(value):
        return False
    value = str(value).strip()
    if len(value) == 0:
        return False
    return bool(regex.search(r"\S", value, flags=regex.UNICODE))


def is_timestamp(value: Any) -> bool:
    """
    Checks if a value is a valid timestamp.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_timestamp("2019-10-23T13:45:12Z")
        # result == True
        ```

        ```python
        result = xmllib.is_timestamp("not a time stamp")
        # result == False
        ```
    """
    validation_regex = r"^\d{4}-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d(\.\d{1,12})?(Z|[+-][0-1]\d:[0-5]\d)$"
    return bool(regex.search(validation_regex, str(value)))


def is_dsp_iri(value: Any) -> bool:
    """
    Check if a value is a valid internal DSP IRI.

    Args:
        value: IRI

    Returns:
        True if it is valid, else false

    Examples:
        ```python
        result = xmllib.is_dsp_iri("http://rdfh.ch/4123/54SYvWF0QUW6a")
        # result == True
        ```

        ```python
        result = xmllib.is_dsp_iri("http://dbpedia.org/resource/Internationalized_Resource_Identifier")
        # result == False
        ```
    """
    return bool(regex.search(r"^http://rdfh\.ch/\d{4}/", str(value)))


def is_dsp_ark(value: Any) -> bool:
    """
    Checks if a value is a valid ARK.

    Args:
        value: ARK

    Returns:
        True if it is valid, else false

    Examples:
        ```python
        result = xmllib.is_dsp_ark("ark:/72163/4123-31ec6eab334-a.2022829")
        # result == True
        ```

        ```python
        result = xmllib.is_dsp_ark("http://rdfh.ch/4123/54SYvWF0QUW6a")
        # result == False
        ```
    """
    return bool(regex.search(r"^ark:/", str(value)))


def check_richtext_syntax(richtext: str) -> None:
    """
    DSP richtexts must be convertible into valid XML.
    This checker escapes the reserved characters `<`, `>` and `&`,
    but only if they are not part of a standard standoff tag or escape sequence.
    Then, it tries to parse the resulting XML.

    Note: Only DSP standard standoff tags are allowed in richtexts. They are documented
    [here](https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/).

    Args:
        richtext: richtext to check

    Warns:
        DspToolsUserWarning: if the input contains XML syntax problems
    """
    escaped_text = _escape_reserved_chars(richtext)
    # transform named entities (=character references) to numeric entities, e.g. &nbsp; -> &#160;
    num_ent = numeric_entities(escaped_text)
    pseudo_xml = f"<text>{num_ent}</text>"
    try:
        _ = etree.fromstring(pseudo_xml)
    except etree.XMLSyntaxError as err:
        prob = IllegalTagProblem(orig_err_msg=err.msg, pseudo_xml=pseudo_xml)
        warnings.warn(DspToolsUserWarning(prob.execute_error_protocol()))


def _escape_reserved_chars(richtext: str) -> str:
    allowed_tags = [  # defined at https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/
        "a( [^>]+)?",  # <a> is the only tag that can have attributes
        "p",
        "em",
        "strong",
        "u",
        "sub",
        "sup",
        "strike",
        "h1",
        "ol",
        "ul",
        "li",
        "tbody",
        "table",
        "tr",
        "td",
        "br",
        "hr",
        "pre",
        "cite",
        "blockquote",
        "code",
    ]
    allowed_tags_regex = "|".join(allowed_tags)
    lookahead = rf"(?!/?({allowed_tags_regex})/?>)"
    illegal_lt = rf"<{lookahead}"
    lookbehind = rf"(?<!</?({allowed_tags_regex})/?)"
    illegal_gt = rf"{lookbehind}>"
    illegal_amp = r"&(?![#a-zA-Z0-9]+;)"
    richtext = regex.sub(illegal_lt, "&lt;", richtext or "")
    richtext = regex.sub(illegal_gt, "&gt;", richtext)
    richtext = regex.sub(illegal_amp, "&amp;", richtext)
    return richtext



import pytest
from lxml import etree

from dsp_tools.commands.ingest_xmlupload.create_resources.apply_ingest_id import replace_filepath_with_internal_filename


class TestReplaceBitstreamPaths:
    def test_all_good(self) -> None:
        xml = """
        <knora>
            <resource label="Fluffy Image" restype=":Image2D" id="Fluffy1" permissions="open">
                <bitstream permissions="open">images/Fluffy.jpg</bitstream>
            </resource>
        </knora>
        """
        root = etree.ElementTree(etree.fromstring(xml)).getroot()
        reference_dict = {"images/Fluffy.jpg": "fluffy_id"}
        res_root, ingest_info = replace_filepath_with_internal_filename(root, reference_dict)
        assert not ingest_info.mediafiles_no_id
        assert not ingest_info.unused_mediafiles
        res_bitstream = res_root[0][0]
        assert res_bitstream.text == "fluffy_id"
        assert res_bitstream.attrib["permissions"] == "open"
        assert res_bitstream.tag == "bitstream"

    def test_extra_paths(self) -> None:
        xml = """
        <knora>
            <resource label="Fluffy Image" restype=":Image2D" id="Fluffy1" permissions="open">
                <bitstream permissions="open">images/Fluffy.jpg</bitstream>
            </resource>
        </knora>
        """
        root = etree.ElementTree(etree.fromstring(xml)).getroot()
        reference_dict = {"images/Fluffy.jpg": "fluffy_id", "extra_media": "extra_id"}
        res_root, ingest_info = replace_filepath_with_internal_filename(root, reference_dict)
        assert not ingest_info.mediafiles_no_id
        assert ingest_info.unused_mediafiles == ["extra_media"]
        res_bitstream = res_root[0][0]
        assert res_bitstream.text == "fluffy_id"
        assert res_bitstream.attrib["permissions"] == "open"
        assert res_bitstream.tag == "bitstream"

    def test_missing_paths(self) -> None:
        xml = """
        <knora>
            <resource label="Fluffy Image" restype=":Image2D" id="Fluffy1" permissions="open">
                <bitstream permissions="open">images/Fluffy.jpg</bitstream>
            </resource>
        </knora>
        """
        root = etree.ElementTree(etree.fromstring(xml)).getroot()
        reference_dict = {"extra_media": "extra_id"}
        res_root, ingest_info = replace_filepath_with_internal_filename(root, reference_dict)
        assert ingest_info.mediafiles_no_id == [("Fluffy1", "images/Fluffy.jpg")]
        assert ingest_info.unused_mediafiles == ["extra_media"]
        res_bitstream = res_root[0][0]
        assert res_bitstream.text == "images/Fluffy.jpg"
        assert res_bitstream.attrib["permissions"] == "open"
        assert res_bitstream.tag == "bitstream"

    def test_absolute_path(self) -> None:
        """
        Ingest strips the leading slash of absolute paths,
        so that the mapping contains the path without leading slash.
        """
        path_in_xml = "/Volumes/images/Fluffy.jpg"
        path_in_mapping = "Volumes/images/Fluffy.jpg"
        xml = f"""
        <knora>
            <resource label="label" restype=":restype" id="id"><bitstream>{path_in_xml}</bitstream></resource>
        </knora>
        """
        root = etree.ElementTree(etree.fromstring(xml)).getroot()
        reference_dict = {path_in_mapping: "fluffy_id"}
        res_root, ingest_info = replace_filepath_with_internal_filename(root, reference_dict)
        assert not ingest_info.mediafiles_no_id
        assert not ingest_info.unused_mediafiles
        res_bitstream = res_root[0][0]
        assert res_bitstream.text == "fluffy_id"
        assert res_bitstream.tag == "bitstream"


if __name__ == "__main__":
    pytest.main([__file__])



import warnings
from dataclasses import dataclass

from typing_extensions import deprecated

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.xmllib.value_checkers import is_dsp_ark
from dsp_tools.xmllib.value_checkers import is_dsp_iri
from dsp_tools.xmllib.value_checkers import is_timestamp

LIST_SEPARATOR = "\n    - "


@dataclass
class MigrationMetadata:
    """
    Metadata originating from a SALSAH migration.
    This will be deprecated in the future when all projects are migrated.
    """

    creation_date: str | None
    iri: str | None
    ark: str | None
    res_id: str

    @deprecated("This is for salsah migration only and will be deleted in future releases.")
    def __post_init__(self) -> None:
        msg_list = []
        if self.creation_date and not is_timestamp(self.creation_date):
            msg_list.append(f"The value for creation date is not a valid timestamp: {self.creation_date}")
        if self.iri and not is_dsp_iri(self.iri):
            msg_list.append(f"The provided IRI is not valid: {self.iri}")
        if self.ark and not is_dsp_ark(self.ark):
            msg_list.append(f"The provided ARK is not valid: {self.ark}")
        if msg_list:
            msg = (
                f"The migration metadata of the resource with the ID '{self.res_id}' has the following problem(s):"
                f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(msg_list)}"
            )
            warnings.warn(DspToolsUserWarning(msg))

    def as_attrib(self) -> dict[str, str]:
        attrib_dict = {}
        if self.creation_date:
            attrib_dict["creation_date"] = self.creation_date
        if self.iri:
            attrib_dict["iri"] = self.iri
        if self.ark:
            attrib_dict["ark"] = self.ark
        if not attrib_dict:
            msg = (
                f"The metadata of the resource with the ID '{self.res_id}' does not contain any values. "
                f"Please check if an error occurred."
            )
            warnings.warn(DspToolsUserWarning(msg))
        return attrib_dict



import pytest
from rdflib import RDF
from rdflib import RDFS
from rdflib import XSD
from rdflib import BNode
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xmlupload.make_rdf_graph.constants import KNORA_API
from dsp_tools.commands.xmlupload.make_rdf_graph.make_resource_and_values import _make_migration_metadata
from dsp_tools.commands.xmlupload.make_rdf_graph.make_resource_and_values import _make_resource
from dsp_tools.commands.xmlupload.models.intermediary.resource import IntermediaryResource
from dsp_tools.commands.xmlupload.models.intermediary.resource import MigrationMetadata
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.models.datetimestamp import DateTimeStamp

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")
RESOURCE_TYPE = URIRef("http://0.0.0.0:3333/ontology/9999/onto/v2#TestResource")
LABEL = Literal("Resource Label", datatype=XSD.string)
PROJECT_IRI = URIRef("http://rdfh.ch/9999/project")


@pytest.fixture
def migration_metadata() -> MigrationMetadata:
    return MigrationMetadata(
        "http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA", DateTimeStamp("1999-12-31T23:59:59.9999999+01:00")
    )


def test_make_resource_mandatory_only() -> None:
    res = IntermediaryResource(
        res_id="resource_id",
        type_iri="http://0.0.0.0:3333/ontology/9999/onto/v2#TestResource",
        label="Resource Label",
        permissions=None,
        values=[],
        file_value=None,
        iiif_uri=None,
        migration_metadata=None,
    )
    res_bn = BNode()
    res_graph = _make_resource(res, res_bn, PROJECT_IRI)
    assert len(res_graph) == 3
    assert next(res_graph.objects(res_bn, RDF.type)) == RESOURCE_TYPE
    assert next(res_graph.objects(res_bn, RDFS.label)) == LABEL
    assert next(res_graph.objects(res_bn, KNORA_API.attachedToProject)) == PROJECT_IRI


def test_make_resource_permissions() -> None:
    res = IntermediaryResource(
        res_id="resource_id",
        type_iri="http://0.0.0.0:3333/ontology/9999/onto/v2#TestResource",
        label="Resource Label",
        permissions=Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]}),
        values=[],
        file_value=None,
        iiif_uri=None,
        migration_metadata=None,
    )
    res_bn = BNode()
    res_graph = _make_resource(res, res_bn, PROJECT_IRI)
    assert len(res_graph) == 4
    assert next(res_graph.objects(res_bn, RDF.type)) == RESOURCE_TYPE
    assert next(res_graph.objects(res_bn, RDFS.label)) == LABEL
    assert next(res_graph.objects(res_bn, KNORA_API.attachedToProject)) == PROJECT_IRI
    permissions = next(res_graph.objects(res_bn, KNORA_API.hasPermissions))
    assert permissions == Literal("CR knora-admin:ProjectAdmin", datatype=XSD.string)


def test_make_resource_migration_metadata() -> None:
    res = IntermediaryResource(
        res_id="resource_id",
        type_iri="http://0.0.0.0:3333/ontology/9999/onto/v2#TestResource",
        label="Resource Label",
        permissions=None,
        values=[],
        file_value=None,
        iiif_uri=None,
        migration_metadata=None,
    )
    res_iri = URIRef("http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA")
    res_graph = _make_resource(res, res_iri, PROJECT_IRI)
    assert len(res_graph) == 3
    assert next(res_graph.objects(res_iri, RDF.type)) == RESOURCE_TYPE
    assert next(res_graph.objects(res_iri, RDFS.label)) == LABEL
    assert next(res_graph.objects(res_iri, KNORA_API.attachedToProject)) == PROJECT_IRI


def test_make_migration_metadata_with_date(migration_metadata: MigrationMetadata) -> None:
    res_iri = URIRef("http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA")
    graph = _make_migration_metadata(migration_metadata, res_iri)
    assert len(graph) == 1
    creation_date = next(graph.objects(res_iri, KNORA_API.creationDate))
    assert creation_date == Literal("1999-12-31T23:59:59.9999999+01:00", datatype=XSD.dateTimeStamp)


def test_make_migration_metadata_no_date() -> None:
    res_iri = URIRef("http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA")
    migration_data = MigrationMetadata("http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA", None)
    graph = _make_migration_metadata(migration_data, res_iri)
    assert len(graph) == 0



import urllib
from pathlib import Path

import pytest
from requests_mock import Mocker

from dsp_tools.commands.xmlupload.models.ingest import DspIngestClientLive
from dsp_tools.models.exceptions import BadCredentialsError
from dsp_tools.models.exceptions import PermanentConnectionError
from test.integration.commands.xmlupload.authentication_client_mock import AuthenticationClientMockBase


@pytest.fixture
def dsp_ingest_url() -> str:
    return "https://example.com"


@pytest.fixture
def shortcode() -> str:
    return "0001"


@pytest.fixture
def ingest_client(dsp_ingest_url: str, shortcode: str) -> DspIngestClientLive:
    return DspIngestClientLive(dsp_ingest_url, AuthenticationClientMockBase(), shortcode, ".")


@pytest.fixture
def tmp_file(tmp_path: Path) -> Path:
    return tmp_path / "éèêëàâæç îïôœùûüÿ.xml"


def _make_url(dsp_ingest_url: str, shortcode: str, file: Path) -> str:
    filename = urllib.parse.quote(file.name)
    return f"{dsp_ingest_url}/projects/{shortcode}/assets/ingest/{filename}"


def test_ingest_success(
    dsp_ingest_url: str, ingest_client: DspIngestClientLive, requests_mock: Mocker, shortcode: str, tmp_file: Path
) -> None:
    tmp_file.write_text("<xml></xml>")
    requests_mock.post(_make_url(dsp_ingest_url, shortcode, tmp_file), json={"internalFilename": tmp_file.name})
    res = ingest_client._ingest(tmp_file)
    assert res.internal_filename == tmp_file.name


def test_ingest_failure_when_file_not_found(ingest_client: DspIngestClientLive, tmp_file: Path) -> None:
    with pytest.raises(FileNotFoundError):
        ingest_client._ingest(tmp_file)


def test_ingest_failure_when_authentication_error(
    dsp_ingest_url: str, ingest_client: DspIngestClientLive, requests_mock: Mocker, shortcode: str, tmp_file: Path
) -> None:
    tmp_file.write_text("<xml></xml>")
    requests_mock.post(_make_url(dsp_ingest_url, shortcode, tmp_file), status_code=401)
    with pytest.raises(BadCredentialsError):
        ingest_client._ingest(tmp_file)


def test_ingest_failure_when_other_error(
    dsp_ingest_url: str, ingest_client: DspIngestClientLive, requests_mock: Mocker, shortcode: str, tmp_file: Path
) -> None:
    tmp_file.write_text("<xml></xml>")
    requests_mock.post(_make_url(dsp_ingest_url, shortcode, tmp_file), status_code=500)
    with pytest.raises(PermanentConnectionError):
        ingest_client._ingest(tmp_file)



from __future__ import annotations

from dataclasses import dataclass

from rdflib import Graph
from rdflib.term import Node


@dataclass
class SHACLGraphs:
    cardinality: Graph
    content: Graph


@dataclass
class RDFGraphs:
    data: Graph
    ontos: Graph
    cardinality_shapes: Graph
    content_shapes: Graph
    knora_api: Graph

    def get_data_and_onto_str(self) -> str:
        g = self.data + self.ontos + self.knora_api
        return g.serialize(format="ttl")

    def get_data_str(self) -> str:
        return self.data.serialize(format="ttl")

    def get_cardinality_shacl_and_onto_str(self) -> str:
        g = self.cardinality_shapes + self.ontos + self.knora_api
        return g.serialize(format="ttl")

    def get_content_shacl_and_onto_str(self) -> str:
        g = self.content_shapes + self.ontos + self.knora_api
        return g.serialize(format="ttl")


@dataclass
class ValidationReportGraphs:
    conforms: bool
    validation_graph: Graph
    shacl_graph: Graph
    onto_graph: Graph
    data_graph: Graph


@dataclass
class QueryInfo:
    validation_bn: Node
    focus_iri: Node
    focus_rdf_type: Node


@dataclass
class UnexpectedComponent:
    component_type: str


@dataclass
class ValidationResultBaseInfo:
    result_bn: Node
    source_constraint_component: Node
    resource_iri: Node
    res_class_type: Node
    result_path: Node
    detail: DetailBaseInfo | None = None


@dataclass
class DetailBaseInfo:
    detail_bn: Node
    source_constraint_component: Node


@dataclass
class ValidationResult:
    res_iri: Node
    res_class: Node
    property: Node


@dataclass
class ResultUniqueValueViolation(ValidationResult):
    actual_value: Node


@dataclass
class ResultValueTypeViolation(ValidationResult):
    results_message: str
    actual_value_type: Node


@dataclass
class ResultPatternViolation(ValidationResult):
    results_message: str
    actual_value: str


@dataclass
class ResultGenericViolation(ValidationResult):
    results_message: str
    actual_value: str


@dataclass
class ResultLinkTargetViolation(ValidationResult):
    expected_type: Node
    target_iri: Node
    target_resource_type: Node | None


@dataclass
class ResultMaxCardinalityViolation(ValidationResult):
    results_message: str


@dataclass
class ResultMinCardinalityViolation(ValidationResult):
    results_message: str


@dataclass
class ResultNonExistentCardinalityViolation(ValidationResult): ...


@dataclass
class ResultFileValueNotAllowedViolation(ValidationResult): ...


@dataclass
class ResultFileValueViolation(ValidationResult):
    results_message: str


@dataclass
class ReformattedIRI:
    res_id: str
    res_type: str
    prop_name: str
